/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Humberto Longo                                    */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

///////////////////////////////////////////////////////////////////////////////
// FILE : CarpIO.h                                                           //
//---------------------------------------------------------------------------//
// Class that read the Carp instance data                                    //
//---------------------------------------------------------------------------//
// AUTHOR : Humberto Jos Longo                                              //
//---------------------------------------------------------------------------//
// CREATED ON : 27/02/2004                                                   //
// LAST MODIFIED :                                                           //
///////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "Constants.h"
#include "FloydWarshall.h"

//---------------------------------------------------------------------------//

#ifndef CARPIO_H_
#define CARPIO_H_

//---------------------------------------------------------------------------//

class CarpIO
{

  public:

    /**
     * Default constructor
     */
    CarpIO();

    /**
     * Free memory allocated
     */
    ~CarpIO();

    /**
     * Read the carp instance data
     */
    void readData( char* _fname );

    /**
     * Converts a carp instance to a cvrp instance (Pearn's algorithm)
     */
    void toCvrpByPearn();

    /**
     * Converts a carp instance to a cvrp instance (Longo's thesis, Uchoa's algorithm)
     */
    void toCvrpByLongo();

    /**
     * Converts a carp instance to a cvrp instance (Longo's thesis, Uchoa's algorithm)
     */
    void toCvrpByUchoa( int option );

    /**
     * Write the cvrp instance file
     */
    void writeCvrpFile();

    /**
     * Write the cvrp distance table
     */
    void writeCvrpDist();

    /**
     * Write the cvrp demand table
     */
    void writeCvrpDem();

    /**
     * Write the cvrp distance table computed by Pearn's algorithm
     */
    void writePearnDist();

    /**
     * Write the cvrp distance table computed by Longo's thesis algorithm
     */
    void writeLongoDist();

    /**
     * Write the cvrp distance table computed by Uchoa's algorithm
     */
    void writeUchoaDist();

    /** Number of vertex - carp graph */
    int nNodes;

    /** Number of arcs - carp graph */
    int nArcs;

    /** Number of required arcs - carp graph */
    int reqArcs;

    /** Number of vertex - cvrp graph */
    int cvrpSize;

    /** Number of vehicles */
    int nVehicles;

    /** Vehicles capacity */
    int capacity;

    /** Carp solution lower bound */
    int carpLB;

    /** Carp solution upper bound */
    int carpUB;

    /** Size of the cvrp to carp convert table */
    int nodeTabSize;

    /** Carp edge table **/
    int** edgeTab;

    /** Cvrp to Carp convert table **/
    int** nodeTab;

    /** Carp arc demand matrix */
    double*  carpDem;

    /** Cvrp node demand matrix */
    double*  cvrpDem;

    /** Carp arc distance matrix */
    double** carpDist;

    /** Cvrp node distance matrix */
    double** cvrpDist;

    /** Class to compute shortest paths */
    FloydWarshall fw;

  private:

    char fname[256];

    double sumDem;

    double sumDist;
};

//---------------------------------------------------------------------------//

#endif
