/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Humberto Longo                                    */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

///////////////////////////////////////////////////////////////////////////////
// FILE : CFloydWarshall.H                                                   //
//---------------------------------------------------------------------------//
// Class to compute shortest paths using the Floyd-Warshall algorithm        //
//---------------------------------------------------------------------------//
// AUTHOR : Humberto Jos Longo                                              //
//---------------------------------------------------------------------------//
// CREATED ON : 26/02/2004                                                   //
// LAST MODIFIED :                                                           //
///////////////////////////////////////////////////////////////////////////////

//---------------------------------------------------------------------------//

#include <stdlib.h>

//---------------------------------------------------------------------------//

#ifndef CFLOYDWARSHALL_H_
#define CFLOYDWARSHALL_H_

//---------------------------------------------------------------------------//

class FloydWarshall
{

  public:

    /**
     * Default constructor
     */
    FloydWarshall();

    /**
     * Default constructor
     */
    FloydWarshall( int _nNodes );

    /**
     * Free memory allocated
     */
    ~FloydWarshall();

    /**
     * Alloc data structures
     * @param graphSize number of nodes 
     */
    void allocDataStructs( int graphSize );

    /**
     * Compute the shortest paths
     * @param graphSize number of nodes 
     * @param edgeCost edge cost matrix
     */
    void shortestPaths( int graphSize, double** edgeCost );


    /** Store the shortest paths among nodes */
    int** minPath;

    /** Store the shortest distance among nodes */
    double** minDist;


  private:

    /** Number of vertex */
    int nNodes;

    /** Work variable to store the shortest paths between nodes */
    int*** path;

    /** Work variable to store the shortest distance between nodes */
    double*** dist;
};

//---------------------------------------------------------------------------//

#endif
