/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CBBNodeList.h                    //
////////////////////////////////////////////
// Class that a list of CBBNode objects.  //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CBBNodeList;

//---------------------------------------------------------------------------//

#include "CBBNode.h"
#include "Constants.h"

//---------------------------------------------------------------------------//

#ifndef CBBNODELIST_H_
#define CBBNODELIST_H_

//---------------------------------------------------------------------------//

/** 
 * Class that represents an element of
 * the list of nodes
 */
class CBBNodeListElem
{
public:

   /** 
    * Default constructor
    */
   CBBNodeListElem();

   /** 
    * Free memory allocated
    */
   ~CBBNodeListElem();

   /** Node of the branch tree */
   CBBNode *node;

   /** Pointer to the next element of the list */
   CBBNodeListElem *next;

   /** Pointer to the previous element of the list */
   CBBNodeListElem *prev;
};

//---------------------------------------------------------------------------//

/**
 * Class that represents the list of nodes
 */
class CBBNodeList
{
public:

   /** 
    * Default constructor
    */
   CBBNodeList();

   /** 
    * Free memory allocated
    */
   ~CBBNodeList();

   /**
    * Add the node in the last position of the list
    * @param node pointer to the node to be inserted
    */
   void addLast(CBBNode *node);

   /**
    * Remove the node from the list
    * @param node the pointer to the node to be removed
    */
   void remove(CBBNode *node);

   /**
    * Empty the node list
    */
   void empty();

   /** Pointer to the first element of the list */
   CBBNodeListElem *first;

   /** Pointer to the last element of the list */
   CBBNodeListElem *last;

   /** Size of the list */
   int size;

};

//---------------------------------------------------------------------------//

#endif


