/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CColPool.h                       //
////////////////////////////////////////////
// Class that represents a pool of cols.  //
//////////////////////////////////////////// 

//---------------------------------------------------------------------------//

class CColPool;

//---------------------------------------------------------------------------//

#ifndef CCOLPOOL_H_
#define CCOLPOOL_H_

//---------------------------------------------------------------------------//

#include "Constants.h"
#include "CColumn.h"

//---------------------------------------------------------------------------//

/**
 * Class that represents an item of a column pool
 */
class CColPoolElem
{

public:

   /** 
    * Default constructor
    */
   CColPoolElem();

   /** 
    * Free the memory allocated
    */
   ~CColPoolElem();

   /** Item of the list (a column) */
   CColumn *column;

   /** Pointer to the next element */
   CColPoolElem* next;

   /** Pointer to the previous element */
   CColPoolElem* prev;

};

//---------------------------------------------------------------------------//

/**
 * Class that represents a pool of columns
 */
class CColPool
{

public: 

   /** 
    * Default constructor
    */
   CColPool();

   /** 
    * Free the memory allocated
    */
   ~CColPool();

   /** 
    * Returns a pointer to the first element of the pool
    * @return a pointer to the first element of the pool
    */
   inline CColPoolElem* getFirst(){ return first; }

   /** 
    * Returns a pointer to the last element of the pool
    * @return a pointer to the last element of the pool
    */
   inline CColPoolElem* getLast(){ return last; };

   /**
    * Returns the number of diferent columns in the pool
    * @return the number of diferent columns in the pool
    */
   inline int getSize(){ return nCols; }

   /**
    * Clean the pool eliminating all the unused columns
    * @return the number of columns deleted from the pool
    */
   int clean();

   /**
    * Remove all the columns from the pool
    */
   void empty();

   /**
    * Add the column to the pool (making a copy of the column)
    * @param aCol the column to be added to the pool
    * @return a pointer to the column added
    */
   CColumn* addColumn( const CColumn &aCol );

   /**
    * Search for the column in the pool
    * @param aCol the column to be searched
    * @return true, if the column is in the pool; false, otherwise
    */
   bool isInPool( const CColumn &aCol );


private:

   /** Pointer to the first column of the pool */
   CColPoolElem *first;

   /** Pointer to the last column of the pool */
   CColPoolElem *last;

   /** Number of columns in the pool */
   int nCols;

};

//---------------------------------------------------------------------------//

#endif
