/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CCutPool.h                       //
////////////////////////////////////////////
// Class that represents a pool of cuts.  //
////////////////////////////////////////////

//---------------------------------------------------------------------------//

class CCutPool;

//---------------------------------------------------------------------------//

#ifndef CCUTPOOL_H_
#define CCUTPOOL_H_

//---------------------------------------------------------------------------//

#include "Constants.h"
#include "CRow.h"

//---------------------------------------------------------------------------//

/**
 * Class that represents an item of a cut pool
 * (a row)
 */
class CCutPoolElem
{

public:

   /**  
    * Default constructor
    */
   CCutPoolElem();

   /** 
    * Free the memory allocated
    */
   ~CCutPoolElem();

   /** Item of the list (a row) */
   CRow *row;

   /** Pointer to the next element */
   CCutPoolElem* next;

   /** Pointer to the previous element */
   CCutPoolElem* prev;

};

//---------------------------------------------------------------------------//

/**
 * Class that represents a pool of cuts (rows)
 */
class CCutPool
{

public: 

   /** 
    * Default constructor
    */
   CCutPool();

   /** 
    * Free the memory allocated
    */
   ~CCutPool();

   /** 
    * Returns a pointer to the first element of the pool
    * @return a pointer to the first element of the pool
    */
   inline CCutPoolElem* getFirst(){ return first; }

   /** 
    * Returns a pointer to the last element of the pool
    * @return a pointer to the last element of the pool
    */
   inline CCutPoolElem* getLast(){ return last; };

   /**
    * Returns the number of diferent cuts in the pool
    * @return the number of diferent cuts in the pool
    */
   inline int getSize(){ return nCuts; }

   /**
    * Clean the pool eliminating all the unused cuts
    * @return the number of cuts deleted from the pool
    */
   int clean();

   /**
    * Remove all the cuts from the pool
    */
   void empty();

   /**
    * Add the cut to the pool (making a copy of the cut)
    * @param aRow the cut (row) to be added to the pool
    * @return a pointer to the row added
    */
   CRow* addRow( const CRow &aRow );

   /**
    * Search for the cut in the pool
    * @param aRow the cut (row) to be searched
    * @return true, if the cut is in the pool; false, otherwise
    */
   bool isInPool( const CRow &aRow );


private:

   /** Pointer to the first cut of the pool */
   CCutPoolElem *first;

   /** Pointer to the last cut of the pool */
   CCutPoolElem *last;

   /** Number of cuts in the pool */
   int nCuts;

};

//---------------------------------------------------------------------------//

#endif
