/****************************************************************************/
/*                                                                          */
/*  This file is distributed as part of the BCP_VRP package                 */
/*                                                                          */
/*  (c) Copyright 2006 by Ricardo Fukasawa, Humberto Longo,                 */
/*  Marcus Poggi de Aragao, Marcelo Reis and Eduardo Uchoa                  */
/*                                                                          */
/*  Permission is granted for academic research use.  For other uses,       */
/*  contact the authors for licensing options.                              */
/*                                                                          */
/*  Use at your own risk.  We make no guarantees about the                  */
/*  correctness or usefulness of this code.                                 */
/*                                                                          */
/****************************************************************************/

////////////////////////////////////////////
// File: CBBNodeList.cpp                  //
////////////////////////////////////////////
// Class that a list of CBBNode objects.  //
////////////////////////////////////////////

#include "CBBNodeList.h"

//---------------------------------------------------------------------------//

CBBNodeListElem::CBBNodeListElem()
{
   next = NULL;
   prev = NULL;
   node = NULL;
}

//---------------------------------------------------------------------------//

CBBNodeListElem::~CBBNodeListElem()
{
   next = NULL;
   prev = NULL;
}

//---------------------------------------------------------------------------//

CBBNodeList::CBBNodeList()
{
   first = NULL;
   last = NULL;
   size = 0;
}

//---------------------------------------------------------------------------//

CBBNodeList::~CBBNodeList()
{
   CBBNodeListElem *aux,*aux2; // auxiliary elements

   aux = first;

   while (aux)
   {
      aux2 = aux->next;

      delete aux;
      size--;

      aux = aux2;
   }

   first = NULL;
   last = NULL;

   if ( size )
      throw(MEMNOTDEALLOCEXCEPTION);
}

//---------------------------------------------------------------------------//

void CBBNodeList::addLast(CBBNode *node)
{
   CBBNodeListElem *aux; // auxiliary list element

   aux = NULL;

   // Allocate node
   aux = new CBBNodeListElem;

   if ( !aux )
      throw(LOWMEMORYEXCEPTION);

   aux->next = NULL;
   aux->prev = NULL;
   aux->node = node;

   if ( first )
   {
      aux->prev = last;
      last->next = aux;
      last = aux;
      size++;
   }
   else
   {
      first = aux;
      last = aux;
      size++;
   }
}

//---------------------------------------------------------------------------//

void CBBNodeList::remove(CBBNode *node)
{
   CBBNodeListElem *aux; // auxiliary elements

   if (!first)
      return;

   if ( first == last)
   {
      if ( first->node == node )
      {
         delete first;
         first = NULL;
         last = NULL;
         size--;
         return;
      }
      else
         return;
   }
   else if ( first->node == node )
   {
      first->next->prev = NULL;
      aux = first;
      first = first->next;
      delete aux;

      size--;
      return;
   }
   else if ( last->node == node )
   {
      last->prev->next = NULL;
      aux = last;
      last = last->prev;
      delete aux;

      size--;
      return;
   }

   aux = first;

   while (aux)
   {
      if ( aux->node == node )
      {
         aux->prev->next = aux->next;
         aux->next->prev = aux->prev;

         delete aux;

         size--;

         return;
      }
      aux = aux->next;
   }
}

//---------------------------------------------------------------------------//

void CBBNodeList::empty()
{
   CBBNodeListElem *aux,*aux2; // auxiliary elements

   aux = first;

   while (aux)
   {
      aux2 = aux->next;

      delete aux;
      size--;

      aux = aux2;
   }

   first = NULL;
   last = NULL;

   if ( size )
      throw(MEMNOTDEALLOCEXCEPTION);
}

//---------------------------------------------------------------------------//


